// Wave2d.java (c) 2004 by Paul Falstad, www.falstad.com

import java.io.InputStream;
import java.awt.*;
import java.awt.image.ImageProducer;
import java.applet.Applet;
import java.util.Vector;
import java.io.File;
import java.util.Random;
import java.util.Arrays;
import java.awt.image.MemoryImageSource;
import java.lang.Math;
import java.awt.event.*;

class Wave2dCanvas extends Canvas {
    Wave2dFrame pg;
    Wave2dCanvas(Wave2dFrame p) {
	pg = p;
    }
    public Dimension getPreferredSize() {
	return new Dimension(300,400);
    }
    public void update(Graphics g) {
	pg.updateWave2d(g);
    }
    public void paint(Graphics g) {
	pg.updateWave2d(g);
    }
};

class Wave2dLayout implements LayoutManager {
    public Wave2dLayout() {}
    public void addLayoutComponent(String name, Component c) {}
    public void removeLayoutComponent(Component c) {}
    public Dimension preferredLayoutSize(Container target) {
	return new Dimension(500, 500);
    }
    public Dimension minimumLayoutSize(Container target) {
	return new Dimension(100,100);
    }
    public void layoutContainer(Container target) {
	Insets insets = target.insets();
	int targetw = target.size().width - insets.left - insets.right;
	int cw = targetw* 7/10;
	int targeth = target.size().height - (insets.top+insets.bottom);
	target.getComponent(0).move(insets.left, insets.top);
	target.getComponent(0).resize(cw, targeth);
	int barwidth = targetw - cw;
	cw += insets.left;
	int i;
	int h = insets.top;
	for (i = 1; i < target.getComponentCount(); i++) {
	    Component m = target.getComponent(i);
	    if (m.isVisible()) {
		Dimension d = m.getPreferredSize();
		if (m instanceof Scrollbar)
		    d.width = barwidth;
		if (m instanceof Label) {
		    h += d.height/5;
		    d.width = barwidth;
		}
		m.move(cw, h);
		m.resize(d.width, d.height);
		h += d.height;
	    }
	}
    }
};

public class Wave2d extends Applet {
    Wave2dFrame ogf;
    void destroyFrame() {
	if (ogf != null)
	    ogf.dispose();
	ogf = null;
    }
    public void init() {
	ogf = new Wave2dFrame(this);
	ogf.init();
    }
    public void destroy() {
	if (ogf != null)
	    ogf.dispose();
	ogf = null;
    }
};

class Wave2dFrame extends Frame
  implements ComponentListener, ActionListener, AdjustmentListener,
             MouseMotionListener, MouseListener, ItemListener {
    
    Thread engine = null;

    Dimension winSize;
    Image dbimage;
    
    Random random;
    int windowWidth = 50;
    int windowHeight = 50;
    int wallWidth;
    
    public String getAppletInfo() {
	return "Wave2d by Paul Falstad";
    }

    Checkbox stoppedCheck;
    Checkbox memoryImageSourceCheck;
    Checkbox intensityCheck;
    Checkbox triChromaticCheck;
    Choice setupChooser;
    Vector setupList;
    Setup setup;
    Button resetTimeButton;
    Scrollbar zoomBar;
    Scrollbar angleBar;
    Scrollbar freqBar;
    Scrollbar resBar;
    Scrollbar speedBar;
    Scrollbar brightnessBar;
    double colorMult;
    Label auxLabels[];
    Scrollbar auxBars[];
    static final int auxBarCount = 5;
    static final double pi = 3.14159265358979323846;
    static final int apertureHeight = 16;
    double colorFunc[][][];
    double apertureR[];
    double apertureI[];
    int dragX, dragY;
    boolean dragging;
    boolean dragClear;
    boolean dragSet;
    boolean recompute;
    double t;
    int pause;
    MemoryImageSource imageSource;
    int pixels[];

    int getrand(int x) {
	int q = random.nextInt();
	if (q < 0) q = -q;
	return q % x;
    }
    Wave2dCanvas cv;
    Wave2d applet;

    Wave2dFrame(Wave2d a) {
	super("Wave2d Applet");
	applet = a;
    }

    public void init() {
	setupList = new Vector();
	Setup s = new SingleSlitSetup();
	while (s != null) {
	    setupList.addElement(s);
	    s = s.createNext();
	}
	String os = System.getProperty("os.name");
	String jv = System.getProperty("java.version");
	int res = 80;
	boolean altRender = false;
	// change settings to speed things up where possible
	if (os.indexOf("Windows") == 0) {
	    res = 120;
	    if (jv.indexOf("1.1") == 0)
		altRender = true;
	}

	setLayout(new Wave2dLayout());
	cv = new Wave2dCanvas(this);
	cv.addComponentListener(this);
	cv.addMouseMotionListener(this);
	cv.addMouseListener(this);
	add(cv);

	setupChooser = new Choice();
	int i;
	for (i = 0; i != setupList.size(); i++)
	    setupChooser.add("Setup: " +
			     ((Setup) setupList.elementAt(i)).getName());
	setup = (Setup) setupList.elementAt(0);
	setupChooser.addItemListener(this);
	add(setupChooser);

	intensityCheck = new Checkbox("Show Intensity", true);
	intensityCheck.addItemListener(this);
	add(intensityCheck);

	stoppedCheck = new Checkbox("Stopped");
	stoppedCheck.addItemListener(this);
	stoppedCheck.disable();
	add(stoppedCheck);
	memoryImageSourceCheck = new Checkbox("Alternate Rendering",
					      altRender);
	memoryImageSourceCheck.addItemListener(this);
	add(memoryImageSourceCheck);

	triChromaticCheck = new Checkbox("Tri-Chromatic", false);
	triChromaticCheck.addItemListener(this);
	add(triChromaticCheck);

	add(resetTimeButton = new Button("Reset Time"));
	resetTimeButton.addActionListener(this);
		
	add(new Label("Incidence Angle", Label.CENTER));
	add(angleBar = new Scrollbar(Scrollbar.HORIZONTAL, 90, 1, 10, 170));
	angleBar.addAdjustmentListener(this);

	add(new Label("Speed", Label.CENTER));
	add(speedBar = new Scrollbar(Scrollbar.HORIZONTAL, 14, 1, 1, 40));
	speedBar.addAdjustmentListener(this);
	speedBar.disable();

	add(new Label("Zoom Out", Label.CENTER));
	add(zoomBar = new Scrollbar(Scrollbar.HORIZONTAL, 10, 1, 10, 200));
	zoomBar.addAdjustmentListener(this);

	add(new Label("Resolution", Label.CENTER));
	add(resBar = new Scrollbar(Scrollbar.HORIZONTAL, res, 5, 120, 600));
	resBar.addAdjustmentListener(this);

	add(new Label("Source Frequency", Label.CENTER));
	add(freqBar = new Scrollbar(Scrollbar.HORIZONTAL,
				     120, 1, 1, 236));
	freqBar.addAdjustmentListener(this);

	add(new Label("Brightness", Label.CENTER));
	add(brightnessBar = new Scrollbar(Scrollbar.HORIZONTAL,
				    27, 1, 1, 1000));
	brightnessBar.addAdjustmentListener(this);

	auxLabels = new Label[auxBarCount];
	auxBars   = new Scrollbar[auxBarCount];
	for (i = 0; i != auxBarCount; i++) {
	    add(auxLabels[i] = new Label("Aux " + (i+1), Label.CENTER));
	    add(auxBars[i] = new Scrollbar(Scrollbar.HORIZONTAL,
					   50, 1, 1, 100));
	    auxBars[i].addAdjustmentListener(this);
	}

	add(new Label("http://www.falstad.com"));

	try {
	    String param = applet.getParameter("PAUSE");
	    if (param != null)
		pause = Integer.parseInt(param);
	} catch (Exception e) { }
	random = new Random();
	setResolution();
	reinit();
	setup = (Setup) setupList.elementAt(0);
	cv.setBackground(Color.black);
	cv.setForeground(Color.lightGray);
	
	resize(750, 600);
	handleResize();
	show();
    }

    void reinit() {
	doSetup();
    }

    void apertureChanged() {
	clearAperture();
	setup.doAperture();
	recompute = true;
    }
    
    void handleResize() {
        Dimension d = winSize = cv.getSize();
	if (winSize.width == 0)
	    return;
	dbimage = createImage(d.width, d.height);
	pixels = new int[d.width*d.height];
	int i;
	for (i = 0; i != d.width*d.height; i++)
	    pixels[i] = 0xFF000000;
	imageSource = new MemoryImageSource(d.width, d.height, pixels, 0,
					    d.width);
    }

    public boolean handleEvent(Event ev) {
        if (ev.id == Event.WINDOW_DESTROY) {
            applet.destroyFrame();
            return true;
        }
        return super.handleEvent(ev);
    }
    
    void centerString(Graphics g, String s, int y) {
	FontMetrics fm = g.getFontMetrics();
        g.drawString(s, (winSize.width-fm.stringWidth(s))/2, y);
    }

    public void paint(Graphics g) {
	cv.repaint();
    }

    boolean calculateNotice;
    public void updateWave2d(Graphics realg) {
	if (winSize == null || winSize.width == 0) {
	    // this works around some weird bug in IE which causes the
	    // applet to not show up properly sometimes.
	    handleResize();
	    return;
	}
        Graphics g = null;

	if (!calculateNotice && recompute) {
	    FontMetrics fm = realg.getFontMetrics();
	    realg.setColor(Color.black);
	    String cs = "Calculating...";
	    realg.fillRect(0, winSize.height-30, 20+fm.stringWidth(cs), 30);
	    realg.setColor(Color.white);
	    realg.drawString(cs, 10, winSize.height-10);
	    cv.repaint(0);
	    calculateNotice = true;
	    return;
	}
	calculateNotice = false;

	boolean mis = memoryImageSourceCheck.getState();
	if (!mis) {
	    g = dbimage.getGraphics();
	    g.setColor(cv.getBackground());
	    g.fillRect(0, 0, winSize.width, winSize.height);
	}
	
	double tadd = 0;
	if (!stoppedCheck.getState()) {
	    int val = speedBar.getValue();
	    tadd = val*.05;
	    // add random crap into the time to avoid aliasing
	    tadd += val*getrand(20) * (.0001091171/4);
	    tadd *= freqBar.getValue() / 34.;
	    t += tadd;
	}
	double trotr = java.lang.Math.cos(t);
	double troti = java.lang.Math.sin(t);
	int i, j;
	
	boolean stopFunc = false;
	if (stoppedCheck.getState())
	    stopFunc = true;
	boolean obstacle = setup instanceof ObstacleSetup;
	if (recompute) {
	    recompute = false;
	    double fm = (angleBar.getValue()-90) * pi/180;
	    double zoom = (zoomBar.getValue()/10.);
	    int apStart = -1, apEnd = -1;
	    int compWidth = (int) (zoom*windowWidth);
	    for (i = 0; i != wallWidth; i++) {
		if (apertureR[i] != 0 || apertureI[i] != 0) {
		    apEnd = i;
		    if (apStart == -1)
			apStart = i;
		}
	    }
	    if (apStart == -1)
		apStart = apEnd = wallWidth/2;
	    int waveStart = wallWidth/2-apEnd;
	    int waveEnd = wallWidth/2+compWidth/2-apStart;
	    int sz = 1;
	    while (sz < compWidth + waveEnd - waveStart)
		sz *= 2;
	    /*System.out.println(apStart + " " + apEnd + " " +
			       waveStart + " " + waveEnd + " " +
			       compWidth + " " + sz + " " + zoom);*/
	    
	    int szm = sz*2-1;
	    double line1[] = new double[sz*2];
	    double line2[] = new double[sz*2];
	    for (i = 0; i != wallWidth; i++)
		if (apertureR[i] != 0 || apertureI[i] != 0) {
		    int ii = (i - wallWidth/2)*2;
		    int i0 = i - wallWidth/2;
		    double a = java.lang.Math.cos(fm*i0);
		    double b = java.lang.Math.sin(fm*i0);
		    line2[ii & szm] = a * apertureR[i] - b * apertureI[i];
		    line2[(ii+1) & szm] =
			a * apertureI[i] + b * apertureR[i];
		}
	    four1(line2, sz, 1);
	    
	    //System.out.print("computing...");
	    int f = 0;
	    int fstart = triChromaticCheck.getState() ? 0 : 1;
	    int fend   = triChromaticCheck.getState() ? 2 : 1;
	    if (intensityCheck.getState())
		t = 1e8;
	    for (f = fstart; f <= fend; f++) {
		double m = freqBar.getValue()/50.;
		if (f == 0)
		    m /= (650/510.); // red, 650 nm (green = 510 nm)
		if (f == 2)
		    m /= (475/510.); // blue, 475 nm
		double m0 = java.lang.Math.sqrt(m);
		double wavelength = 2*pi/m;
		double freq = 1/(2*pi);
		double speed = wavelength * freq;
		// 1.23 =~ sqrt(1 + 1/2)
		if (t < resBar.getValue()*1.23/speed)
		    recompute = true;
		for (j = 0; j != windowHeight; j++) {
		    for (i = 0; i != sz*2; i++)
			line1[i] = 0;
		    int jj = j+1;
		    double jz2 = jj*jj*zoom*zoom;
		    int i2;
		    int sz2 = sz*2;
		    for (i = i2 = 0; i <= waveEnd; i++, i2 += 2) {
			double dist1 = java.lang.Math.sqrt(i*i+jz2);
			double dist2 = dist1*m;
			computeBessel(dist2);
			if (dist1 > t*speed - wavelength/8)
			    bessj0 = bessy0 = 0;
			line1[i2] = bessj0*m0;
			line1[i2+1] = -bessy0*m0;
			if (i > 0) {
			    line1[sz2-i2] = line1[i2];
			    line1[sz2-i2+1] = line1[i2+1];
			}
		    }
		    four1(line1, sz, 1);
		    for (i = 0; i != sz; i++) {
			int ii = i*2;
			double a = line1[ii]*line2[ii] -
			    line1[ii+1]*line2[ii+1];
			double b = line1[ii+1]*line2[ii] +
			    line1[ii]*line2[ii+1];
			line1[ii] = a;
			line1[ii+1] = b;
		    }
		    four1(line1, sz, -1);
		    double qmult = 400./sz;
		    double oaddr = 0, oaddi = 0;
		    if (obstacle) {
			oaddr = java.lang.Math.cos(jj*zoom*m)*800/m0;
			oaddi = -java.lang.Math.sin(jj*zoom*m)*800/m0;
			int ww = (int) (windowWidth * zoom);
			for (i = 0; i != ww; i++) {
			    int ii = i-ww/2;
			    double a = java.lang.Math.cos(fm*ii);
			    double b = java.lang.Math.sin(fm*ii);
			    double aa = (a * oaddr - b * oaddi) / qmult;
			    double bb = (a * oaddi + b * oaddr) / qmult;
			    line1[szm & (ii*2)  ] = aa-line1[szm & (ii*2)];
			    line1[szm & (ii*2)+1] = bb-line1[szm & (ii*2)+1];
			}
		    }
		    for (i = 0; i != windowWidth; i++) {
			double ir1 = (int) ((i-windowWidth/2  )*zoom);
			double ir2 = (int) ((i-windowWidth/2+1)*zoom);
			int ii1 = (int) ir1;
			int ii2 = (int) ir2;
			int iic = ii2-ii1;
			if (intensityCheck.getState())
			    colorFunc[i][j][f] = 0;
			else {
			    colorFunc[i][j][0] = 0;
			    colorFunc[i][j][1] = 0;
			}
			int ii;
			for (ii = ii1; ii <= ii2; ii++) {
			    double q1 = line1[szm & (ii*2  )]*qmult;
			    double q2 = line1[szm & (ii*2+1)]*qmult;
			    double mu = .001;
			    if (ii == ii1)
				mu *= 1-(ir1-ii1);
			    else if (ii == ii2)
				mu *= ir2-ii2;
			    if (intensityCheck.getState())
				colorFunc[i][j][f] += (q1*q1+q2*q2) * mu;
			    else {
				colorFunc[i][j][0] += q1*mu;
				colorFunc[i][j][1] += q2*mu;
			    }
			}
			if (intensityCheck.getState())
			    colorFunc[i][j][f] /= ir2-ir1;
			else {
			    colorFunc[i][j][0] /= ir2-ir1;
			    colorFunc[i][j][1] /= ir2-ir1;
			}
		    }
		}
	    }
	    //System.out.println("done");
	}

	colorMult = 30*java.lang.Math.exp(brightnessBar.getValue()/50.-10);
	if (!intensityCheck.getState())
	    colorMult *= 1.5;
	int ix = 0;
	int k, l;
	int height = winSize.height - apertureHeight;
	for (j = 0; j != windowHeight; j++) {
	    for (i = 0; i != windowWidth; i++) {
		int x = i*winSize.width/windowWidth;
		int y = j*height/windowHeight + apertureHeight;
		int x2 = (i+1)*winSize.width/windowWidth;
		int y2 = (j+1)*height/windowHeight + apertureHeight;
		int colval = 0;
		if (intensityCheck.getState()) {
		    colval = 0xFF000000 +
			(getColorValue(i, j, 0) << 16) |
			(getColorValue(i, j, 1) << 8) |
			(getColorValue(i, j, 2));
		} else {
		    double q1 = colorFunc[i][j][0];
		    double q2 = colorFunc[i][j][1];
		    double q = (q1*trotr - q2*troti)*colorMult;
		    if (q > 0) {
			int qq = (int) (q*255);
			if (qq > 255)
			    qq = 255;
			colval = 0xFF000000 + (qq << 8);
		    } else {
			int qq = (int) (-q*255);
			if (qq > 255)
			    qq = 255;
			colval = 0xFF000000 + (qq << 16);
		    }
		}
		if (mis) {
		    for (k = x; k < x2; k++)
			for (l = y; l < y2; l++)
			    pixels[k+l*winSize.width] = colval;
		} else {
		    g.setColor(new Color(colval));
		    g.fillRect(x, y, x2-x, y2-y);
		}
	    }
	}

	double zoom = (zoomBar.getValue()/10.);
	//System.out.println(zoom + " " + windowWidth + " " + winSize.width);
	for (i = 0; i != windowWidth; i++) {
	    double ir1 = (int) ((i-windowWidth/2  )*zoom);
	    double ir2 = (int) ((i-windowWidth/2+1)*zoom);
	    int ii1 = (int) ir1;
	    int ii2 = (int) ir2;
	    int iic = ii2-ii1;
	    int ii;
	    double funcr = 0, funcb = 0;
	    for (ii = ii1; ii <= ii2; ii++) {
		//System.out.println(i + " " + ii + " " + zoom);
		int ij = ii+wallWidth/2;
		double mu = 1;
		if (ii == ii1)
		    mu *= 1-(ir1-ii1);
		else if (ii == ii2)
		    mu *= ir2-ii2;
		if (ij < 0 || ij >= wallWidth) {
		    funcb += mu;
		    continue;
		}
		funcb += (1-(apertureR[ij]*apertureR[ij] +
			     apertureI[ij]*apertureI[ij]))*mu;
		double ph = java.lang.Math.atan2(apertureI[ij],
						 apertureR[ij])/pi;
		if (ph < 0)
		    funcr += (2+ph)*mu;
		else
		    funcr += ph*mu;
	    }
	    funcb /= ir2-ir1;
	    funcr /= (ir2-ir1)*2;
	    //System.out.println(i + " " + funcb);
	    if (obstacle)
		funcb = 1-funcb;
	    int valb = (int) (funcb*255);
	    int valr = (int) (funcr*255);
	    if (valb < 0)
		valb = 0;
	    if (valb > 255)
		valb = 255;
	    if (valr < 0)
		valr = 0;
	    if (valb > 255)
		valb = 255;
	    int colval = 0xFF000000 + (valr << 16) | valb;
	    int x = i*winSize.width/windowWidth;
	    int x2 = (i+1)*winSize.width/windowWidth;
	    int y = 0;
	    int y2 = apertureHeight;
	    if (mis) {
		for (k = x; k < x2; k++)
		    for (l = y; l < y2; l++)
			pixels[k+l*winSize.width] = colval;
	    } else {
		g.setColor(new Color(colval));
		g.fillRect(x, y, x2-x, y2-y);
	    }
	}
	
	if (mis)
	    dbimage = cv.createImage(imageSource);
	realg.drawImage(dbimage, 0, 0, this);
	if (!intensityCheck.getState() && !stoppedCheck.getState())
	    cv.repaint(pause);
    }

    int getColorValue(int i, int j, int k) {
	int val = (int) (colorFunc[i][j][k] * colorMult);
	if (val > 255)
	    val = 255;
	return val;
    }

    int abs(int x) {
	return x < 0 ? -x : x;
    }

    int sign(int x) {
	return (x < 0) ? -1 : (x == 0) ? 0 : 1;
    }

    public void componentHidden(ComponentEvent e){}
    public void componentMoved(ComponentEvent e){}
    public void componentShown(ComponentEvent e) {
	cv.repaint();
    }

    public void componentResized(ComponentEvent e) {
	handleResize();
	cv.repaint(100);
    }
    public void actionPerformed(ActionEvent e) {
	if (e.getSource() == resetTimeButton) {
	    t = 0;
	    recompute = true;
	    cv.repaint();
	}
    }

    public void adjustmentValueChanged(AdjustmentEvent e) {
	System.out.print(((Scrollbar) e.getSource()).getValue() + "\n");
	if (e.getSource() == resBar && resBar.getValue() != resBarValue)
	    setResolution();
	int i;
	for (i = 0; i != auxBarCount; i++)
	    if (e.getSource() == auxBars[i]) {
		apertureChanged();
		break;
	    }
	if (e.getSource() == zoomBar || e.getSource() == angleBar ||
	    e.getSource() == freqBar)
	    recompute = true;
	setResetTimeButton();
	cv.repaint(pause);
    }

    void setResetTimeButton() {
	// Reset Time button doesn't work well with obstacle or lo freqs
	if (setup instanceof ObstacleSetup || freqBar.getValue() < 8 ||
	    intensityCheck.getState())
	    resetTimeButton.disable();
	else
	    resetTimeButton.enable();
    }
	
    int resBarValue = -1;
    void setResolution() {
	resBarValue = windowWidth = windowHeight = resBar.getValue();
	if ((windowWidth & 1) == 1)
	    windowWidth = windowHeight = resBarValue-1;
	colorFunc = new double[windowWidth][windowHeight][3];
	wallWidth = 512;
	apertureR = new double[wallWidth];
	apertureI = new double[wallWidth];
	System.out.print(windowWidth + " " + windowHeight + "\n");
	apertureChanged();
    }

    void setResolution(int x) {
	resBar.setValue(x);
	setResolution();
    }

    public void mouseDragged(MouseEvent e) {
	dragging = true;
	//edit(e);
	cv.repaint(pause);
    }
    public void mouseMoved(MouseEvent e) {
	if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)
	    return;
	int x = e.getX();
	int y = e.getY();
	dragX = x; dragY = y;
    }
    public void mouseClicked(MouseEvent e) {
    }
    public void mouseEntered(MouseEvent e) {
    }
    public void mouseExited(MouseEvent e) {
    }
    public void mousePressed(MouseEvent e) {
	if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) == 0)
	    return;
	dragging = true;
	//edit(e);
    }
    public void mouseReleased(MouseEvent e) {
	if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) == 0)
	    return;
	dragging = false;
	dragSet = dragClear = false;
	cv.repaint();
    }
    public void itemStateChanged(ItemEvent e) {
	if (e.getItemSelectable() == triChromaticCheck ||
	    e.getItemSelectable() == intensityCheck) {
	    setResolution();
	    recompute = true;
	    cv.repaint();
	    if (intensityCheck.getState()) {
		stoppedCheck.disable();
		speedBar.disable();
		triChromaticCheck.enable();
	    } else {
		stoppedCheck.enable();
		speedBar.enable();
		triChromaticCheck.disable();
	    }
	    setResetTimeButton();
	    return;
	}
	if (e.getItemSelectable() == stoppedCheck) {
	    cv.repaint();
	    return;
	}
	if (e.getItemSelectable() == memoryImageSourceCheck) {
	    dbimage = createImage(winSize.width, winSize.height);
	}
	if (e.getItemSelectable() == setupChooser) {
	    doSetup();
	    cv.repaint(pause);
	}
    }

    void clearAperture() {
	int i;
	for (i = 0; i != wallWidth; i++)
	    apertureR[i] = apertureI[i] = 0;
    }
    
    void doSetup() {
	t = 1e8;
	int i;
	for (i = 0; i != auxBarCount; i++)
	    auxBars[i].setValue(10);
	freqBar.setValue(120);
	zoomBar.setValue(10);
	angleBar.setValue(90);
	setup = (Setup)
	    setupList.elementAt(setupChooser.getSelectedIndex());
	angleBar.enable();
	setResetTimeButton();
	setup.select();
	apertureChanged();
	for (i = 0; i < setup.getAuxBarCount(); i++) {
	    auxLabels[i].show();
	    auxBars[i].show();
	}
	for (; i < auxBarCount; i++) {
	    auxLabels[i].hide();
	    auxBars[i].hide();
	}
	validate();
    }

    abstract class Setup {
	abstract String getName();
	abstract void select();
	abstract void doAperture();
	abstract Setup createNext();
	int getAuxBarCount() { return 2; }
    };

    class SingleSlitSetup extends Setup {
	String getName() { return "Single Slit"; }
	void select() {
	    auxLabels[0].setText("Slit Width");
	    brightnessBar.setValue(440);
	}
	void doAperture() {
	    int x;
	    int w = auxBars[0].getValue();
	    //System.out.println("width = " + w);
	    for (x = 0; x != w; x++)
		apertureR[wallWidth/2 - w/2 + x] = 1;
	}
	int getAuxBarCount() { return 1; }
	Setup createNext() { return new DoubleSlitSetup(); }
    }
    class DoubleSlitSetup extends Setup {
	String getName() { return "Double Slit"; }
	void select() {
	    auxLabels[0].setText("Slit Width");
	    auxLabels[1].setText("Separation");
	    auxLabels[2].setText("Balance");
	    auxLabels[3].setText("Phase Difference");
	    auxBars[2].setValue(50);
	    auxBars[3].setValue(0);
	    brightnessBar.setValue(380);
	}
	void doAperture() {
	    int x;
	    int w = auxBars[0].getValue();
	    int s = auxBars[1].getValue();
	    double bal2 = auxBars[2].getValue() / 100.;
	    double bal1 = 1-bal2;
	    if (bal1 > bal2) {
		bal2 /= bal1;
		bal1 = 1;
	    } else {
		bal1 /= bal2;
		bal2 = 1;
	    }
	    double ph = auxBars[3].getValue() * pi / 50.;
	    double a2r = bal2 * java.lang.Math.cos(ph);
	    double a2i = bal2 * java.lang.Math.sin(ph);
	    for (x = 0; x != w; x++) {
		apertureR[wallWidth/2 - s - x] = bal1;
		apertureR[wallWidth/2 + s + x] = a2r;
		apertureI[wallWidth/2 + s + x] = a2i;
	    }
	}
	int getAuxBarCount() { return 4; }
	Setup createNext() { return new GratingSetup(); }
    }
    class GratingSetup extends Setup {
	String getName() { return "Multiple Slits"; }
	void select() {
	    auxLabels[0].setText("Slit Count");
	    auxLabels[1].setText("Slit Width");
	    auxLabels[2].setText("Separation");
	    brightnessBar.setValue(345);
	}
	int getAuxBarCount() { return 3; }
	void doAperture() {
	    int x;
	    int w = auxBars[1].getValue();
	    int s = auxBars[2].getValue()+3;
	    if (w > s-1)
		w = s-1;
	    int i;
	    int n = auxBars[0].getValue()/5+1;
	    int sub = (s*(n-1)+w)/2;
	    for (x = 0; x != w; x++)
		for (i = 0; i != n; i++)
		    apertureR[wallWidth/2-sub+s*i+x] = 1;
	}
	Setup createNext() { return new ObstacleSetup(); }
    }
    class ObstacleSetup extends Setup {
	String getName() { return "Obstacle"; }
	void select() {
	    auxLabels[0].setText("Width");
	    // waves with an incidence angle don't work for some reason
	    // with obstacles, so disable them
	    angleBar.disable();
	    angleBar.setValue(90);
	    // don't feel like getting Reset Time to work with Obstacle
	    resetTimeButton.disable();
	    brightnessBar.setValue(310);
	}
	int getAuxBarCount() { return 1; }
	void doAperture() {
	    int x;
	    int w = (int) (auxBars[0].getValue()*1.5);
	    for (x = 0; x != w; x++)
		apertureR[wallWidth/2-w/2+x] = 1;
	}
	Setup createNext() { return new ZonePlateEvenSetup(); }
    }
    class ZonePlateEvenSetup extends Setup {
	String getName() { return "Zone Plate (Even)"; }
	int evenOdd = 0;
	boolean phase = false;
	boolean blazed = false;
	void select() {
	    auxLabels[0].setText("Intended Frequency");
	    auxBars[0].setValue(freqBar.getValue() * 100 / 236);
	    auxLabels[1].setText("Focal Length");
	    auxBars[1].setValue(20);
	    auxLabels[2].setText("Plate Width");
	    auxBars[2].setValue(100);
	    brightnessBar.setValue(111);
	}
	int getAuxBarCount() { return 3; }
	void doAperture() {
	    int i;
	    // freqBar ranges from 0 to 4.72, so we match that here
	    double m = auxBars[0].getValue() * .0472;
	    double halfwave = pi/m;
	    int pw = auxBars[2].getValue() * 3;
	    int dy = auxBars[1].getValue()*5;
	    int cx = wallWidth/2;
	    for (i = 0; i != wallWidth; i++) {
		int dx = cx-i;
		if (dx < -pw || dx > pw)
		    continue;
		double dist = java.lang.Math.sqrt(dx*dx+dy*dy);
		dist = (dist-dy);
		if (blazed) {
		    double ph = dist/halfwave * pi;
		    apertureR[i] = java.lang.Math.cos(ph);
		    apertureI[i] = java.lang.Math.sin(ph);
		} else {
		    int zone = (int) (dist / halfwave);
		    apertureR[i] = ((zone & 1) == evenOdd) ? 1 :
			(phase) ? -1 : 0;
		}
	    }
	}
	Setup createNext() { return new ZonePlateOddSetup(); }
    }
    class ZonePlateOddSetup extends ZonePlateEvenSetup {
	ZonePlateOddSetup() { evenOdd = 1; }
	String getName() { return "Zone Plate (Odd)"; }
	Setup createNext() { return new ZonePlatePhaseSetup(); }
    }
    class ZonePlatePhaseSetup extends ZonePlateOddSetup {
	ZonePlatePhaseSetup() { phase = true; }
	String getName() { return "Phase-Reversal Zone Plate"; }
	Setup createNext() { return new ZonePlateBlazedSetup(); }
    }
    class ZonePlateBlazedSetup extends ZonePlateOddSetup {
	ZonePlateBlazedSetup() { blazed = true; }
	String getName() { return "Blazed Zone Plate"; }
	Setup createNext() { return new Hologram1Setup(); }
    }
    class Hologram1Setup extends Setup {
	String getName() { return "Absorption Hologram"; }
	void select() {
	    auxLabels[0].setText("Intended Frequency");
	    auxBars[0].setValue(freqBar.getValue() * 100 / 236);
	    auxLabels[1].setText("X 1");
	    auxLabels[2].setText("Y 1");
	    auxLabels[3].setText("X 2");
	    auxLabels[4].setText("Y 2");
	    auxBars[1].setValue(40);
	    auxBars[2].setValue(15);
	    auxBars[3].setValue(70);
	    auxBars[4].setValue(40);
	    brightnessBar.setValue(285);
	    zoomBar.setValue(15);
	}
	int getAuxBarCount() { return 5; }
	void doAperture() {
	    int i;
	    // freqBar ranges from 0 to 4.72, so we match that here
	    double m = auxBars[0].getValue() * .0472;
	    double halfwave = pi/m;
	    int px1 = auxBars[1].getValue() * 2 - 100;
	    int py1 = auxBars[2].getValue() * 3+5;
	    int px2 = auxBars[3].getValue() * 2 - 100;
	    int py2 = auxBars[4].getValue() * 3+5;
	    int cx = wallWidth/2;
	    int pw = 300;
	    double maxf = 0;
	    double aradd = .75;
	    for (i = 0; i != wallWidth; i++) {
		if (i-cx < -pw || i-cx > pw)
		    continue;
		
		int dx = px1-(i-cx);
		double dist = java.lang.Math.sqrt(dx*dx+py1*py1);
		computeBessel(dist*m);
		double ar = bessj0, ai = bessy0;
		
		dx = px2-(i-cx);
		dist = java.lang.Math.sqrt(dx*dx+py2*py2);
		computeBessel(dist*m);
		ar += bessj0;
		ai += bessy0;

		ar += aradd;
		double q = apertureR[i] = java.lang.Math.sqrt(ar*ar+ai*ai);
		if (q > maxf)
		    maxf = q;
	    }
	    maxf = java.lang.Math.sqrt(maxf);
	    for (i = 0; i != wallWidth; i++)
		apertureR[i] /= maxf;
	}
	Setup createNext() { return new Hologram2Setup(); }
    }
    class Hologram2Setup extends Setup {
	String getName() { return "Phase Hologram"; }
	void select() {
	    auxLabels[0].setText("Intended Frequency");
	    auxBars[0].setValue(freqBar.getValue() * 100 / 236);
	    auxLabels[1].setText("X 1");
	    auxLabels[2].setText("Y 1");
	    auxLabels[3].setText("X 2");
	    auxLabels[4].setText("Y 2");
	    auxBars[1].setValue(40);
	    auxBars[2].setValue(15);
	    auxBars[3].setValue(70);
	    auxBars[4].setValue(40);
	    brightnessBar.setValue(150);
	    zoomBar.setValue(15);
	}
	int getAuxBarCount() { return 5; }
	void doAperture() {
	    int i;
	    // freqBar ranges from 0 to 4.72, so we match that here
	    double m = auxBars[0].getValue() * .0472;
	    double halfwave = pi/m;
	    int px1 = auxBars[1].getValue() * 2 - 100;
	    int py1 = auxBars[2].getValue() * 3+5;
	    int px2 = auxBars[3].getValue() * 2 - 100;
	    int py2 = auxBars[4].getValue() * 3+5;
	    int cx = wallWidth/2;
	    int pw = 300;
	    double maxf = 0;
	    for (i = 0; i != wallWidth; i++) {
		if (i-cx < -pw || i-cx > pw)
		    continue;
		
		int dx = px1-(i-cx);
		double dist = java.lang.Math.sqrt(dx*dx+py1*py1);
		computeBessel(dist*m);
		apertureR[i] = bessj0;
		apertureI[i] = bessy0;
		
		dx = px2-(i-cx);
		dist = java.lang.Math.sqrt(dx*dx+py2*py2);
		computeBessel(dist*m);
		apertureR[i] += bessj0;
		apertureI[i] += bessy0;
		
		double q =
		    apertureR[i]*apertureR[i] +
		    apertureI[i]*apertureI[i];
		if (q > maxf)
		    maxf = q;
	    }
	    maxf = java.lang.Math.sqrt(maxf);
	    for (i = 0; i != wallWidth; i++) {
		apertureR[i] /= maxf;
		apertureI[i] /= maxf;
	    }
	}
	Setup createNext() { return null; }
    }

    // fourier transform
    void four1(double data[],int nn,int isign) {
	int n,mmax,m,j,istep,i;
	double wtemp,wr,wpr,wpi,wi,theta;
	double tempr,tempi;
	
	n=nn << 1;
	j=1;
	for (i=1;i<n;i+=2) {
	    if (j > i) {
		tempr = data[j-1]; data[j-1] = data[i-1]; data[i-1] = tempr;
		tempr = data[j]; data[j] = data[i]; data[i] = tempr;
	    }
	    m=n >> 1;
	    while (m >= 2 && j > m) {
		j -= m;
		m >>= 1;
	    }
	    j += m;
	}
	mmax=2;
	while (n > mmax) {
	    istep=2*mmax;
	    theta=6.28318530717959/(isign*mmax);
	    wtemp=java.lang.Math.sin(0.5*theta);
	    wpr = -2.0*wtemp*wtemp;
	    wpi=java.lang.Math.sin(theta);
	    wr=1.0;
	    wi=0.0;
	    for (m=1;m<mmax;m+=2) {
		for (i=m;i<=n;i+=istep) {
		    j=i+mmax;
		    tempr=wr*data[j-1]-wi*data[j];
		    tempi=wr*data[j]+wi*data[j-1];
		    data[j-1]=data[i-1]-tempr;
		    data[j]=data[i]-tempi;
		    data[i-1] += tempr;
		    data[i] += tempi;
		}
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	    }
	    mmax=istep;
	}
    }

    double bessj0, bessy0;
    void computeBessel(double x) {
	double ax = x,z;
	double xx,y,ans,ans1,ans2;

	if (x < 8.0) {
	    y=x*x;
	    ans1=57568490574.0+y*(-13362590354.0+y*(651619640.7
						    +y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
	    ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
						  +y*(59272.64853+y*(267.8532712+y*1.0))));
	    bessj0 = ans=ans1/ans2;

	    ans1 = -2957821389.0+y*(7062834065.0+y*(-512359803.6
						    +y*(10879881.29+y*(-86327.92757+y*228.4622733))));
	    ans2=40076544269.0+y*(745249964.8+y*(7189466.438
						 +y*(47447.26470+y*(226.1030244+y*1.0))));
	    bessy0=(ans1/ans2)+0.636619772*bessj0*java.lang.Math.log(x);
	    
	} else {
	    z=8.0/ax;
	    y=z*z;
	    xx=ax-0.785398164;
	    ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
					    +y*(-0.2073370639e-5+y*0.2093887211e-6)));
	    ans2 = -0.1562499995e-1+y*(0.1430488765e-3
				+y*(-0.6911147651e-5+y*(0.7621095161e-6
						      -y*0.934935152e-7)));
	    double sax = java.lang.Math.sqrt(0.636619772/ax);
	    double cosxx = java.lang.Math.cos(xx);
	    double sinxx = java.lang.Math.sin(xx);
	    bessj0 = sax * (cosxx*ans1-z*sinxx*ans2);

	    ans2 = -0.1562499995e-1+y*(0.1430488765e-3
				+y*(-0.6911147651e-5+y*(0.7621095161e-6
						      +y*(-0.934945152e-7))));
	    bessy0 = sax*(sinxx*ans1+z*cosxx*ans2);
	    
	}
    }
}
